__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from action_updater.logger import Table
from action_updater.main import get_client


def list_updaters(args, parser, extra, subparser):
    cli = get_client(quiet=args.quiet)
    items = [
        {"title": x.title, "identifier": name, "description": x.description}
        for name, x in cli.updaters.items()
    ]
    table = Table(items)
    table.show()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


import string

import pipelib.pipeline as pipeline
import pipelib.steps as step


def sort_tags(tags):
    """
    Sort a list of string tags, return sorted (first latest) with original version
    """
    # all letters excluded except for v
    letters = "(%s)" % "|".join([x for x in string.ascii_letters if x not in ["v", "V"]])

    # A pipeline to process docker tags
    steps = (
        # No letters except for "v"
        ~step.filters.HasPatterns(filters=[letters]),
        # Scrub commits from version string
        step.filters.CleanCommit(),
        # Parse versions, return sorted ascending, and taking version major.minor.patch into account
        step.container.ContainerTagSort(),
    )
    p = pipeline.Pipeline(steps)
    return p.run(list(tags), unwrap=False)


def sort_major(tags):
    """
    Allow major tags like v3
    """
    # A pipeline to process docker tags
    steps = step.release.MajorTagSort()
    p = pipeline.Pipeline(steps)
    return p.run(list(tags), unwrap=False)



#!/usr/bin/python

# Copyright (C) 2022 Vanessa Sochat.

# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import string

import pytest

from action_updater.main.action import GitHubAction
from action_updater.tests.helpers import get_updaters, here, init_client


@pytest.mark.parametrize("updater_name", get_updaters())
def test_updaters(tmp_path, updater_name):
    """
    Test each updater.
    """
    client = init_client(str(tmp_path))
    updater = client.updaters[updater_name]

    # The updater must have a description and title
    assert updater.description
    assert updater.title

    # The name must be all lowercase, no special characters except for -
    letters = string.ascii_letters + "-"
    for letter in updater.name:
        assert letter in letters

    # A test (before and after) must be defined for each
    before_file = os.path.join(here, "data", f"{updater.name}-before.yaml")
    after_file = os.path.join(here, "data", f"{updater.name}-after.yaml")
    for filename in before_file, after_file:
        assert os.path.exists(filename)

    # Count is 0 before run, and we find changes
    assert updater.count == 0

    # Read into GitHub action
    action = GitHubAction(before_file)

    # Run the whole thing with detect (to print to console)
    result = client.detect(before_file, updaters=[updater.slug])
    result[before_file].render_after() == action.render_after()

    result = updater.detect(action)
    assert result is True
    assert action.has_changes

    # Count is 0 before run, and we find changes
    assert updater.count != 0

    # Now for the after file (shouldn't change until GitHub updates versions, slowly)
    action = GitHubAction(after_file)
    result = updater.detect(action)
    assert result is False



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

__version__ = "0.0.11"
AUTHOR = "Vanessa Sochat"
EMAIL = "vsoch@users.noreply.github.com"
NAME = "action-updater"
PACKAGE_URL = "https://github.com/vsoch/action-updater"
KEYWORDS = "github, actions, updater"
DESCRIPTION = "Update deprecated command, versions, and other for GitHub actions"
LICENSE = "LICENSE"

################################################################################
# Global requirements

INSTALL_REQUIRES = (
    ("pipelib", {"min_version": None}),
    ("jsonschema", {"min_version": None}),
    # Required to maintain comments in files
    ("ruamel.yaml", {"min_version": None}),
    ("rich", {"min_version": None}),
    ("pyyaml", {"min_version": None}),
    ("requests", {"min_version": None}),
    # seems to be an issue for older python
    ("packaging", {"min_version": None}),
)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)

################################################################################
# Submodule Requirements (versions that include database)

INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + TESTS_REQUIRES



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from pygments.styles import get_all_styles

# Derived from https://github.com/softprops/github-actions-schemas/blob/master/workflow.json

schema_url = "http://json-schema.org/draft-07/schema"

workflow_schema = {
    "title": "Github workflow file - https://help.github.com/en/articles/workflow-syntax-for-github-actions",
    "$schema": schema_url,
    "type": "object",
    "required": ["on"],
    "properties": {
        "name": {
            "type": "string",
            "description": "The name of your workflow. GitHub displays the names of your workflows on your repository's actions page. If you omit this field, GitHub sets the name to the workflow's filename.",
        },
        "on": {
            "description": "The name of the GitHub event that triggers the workflow",
            "oneOf": [
                {"$ref": "#/definitions/webhook_event"},
                {"type": "array", "items": {"$ref": "#/definitions/webhook_event"}},
                {
                    "type": "object",
                    "addditionalProperties": {
                        "type": "object",
                        "properties": {
                            "branches": {
                                "description": "list of branches",
                                "type": "array",
                                "items": {"type": "string"},
                            },
                            "tags": {"type": "array", "items": {"type": "string"}},
                            "paths": {"type": "array", "items": {"type": "string"}},
                        },
                    },
                },
                {
                    "type": "object",
                    "properties": {
                        "schedule": {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "properties": {"cron": {"type": "string"}},
                            },
                        }
                    },
                },
            ],
        },
        "jobs": {
            "description": "A workflow run is made up of one or more jobs. Jobs run in parallel by default. To run jobs sequentially, you can define dependencies on other jobs using the jobs.<job_id>.needs keyword.",
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "description": "Each job must have an id to associate with the job. The key job_id is a string and its value is a map of the job's configuration data. You must replace <job_id> with a string that is unique to the jobs object. The <job_id> must start with a letter or _ and contain only alphanumeric characters, -, or _.",
                "required": ["runs-on"],
                "properties": {
                    "name": {
                        "description": "The name of the job displayed on GitHub.",
                        "type": "string",
                    },
                    "needs": {
                        "description": "Identifies any jobs that must complete successfully before this job will run. It can be a string or array of strings. If a job fails, all jobs that need it will also fail unless the jobs use a conditional statement that causes the job to continue.",
                        "oneOf": [
                            {"type": "string"},
                            {"type": "array", "items": {"type": "string"}},
                        ],
                    },
                    "runs-on": {
                        "description": "The type of virtual host machine to run the job on. Each job runs with a fresh instance of the virtual environment specified in by runs-on",
                        "type": "string",
                        "default": "ubuntu-latest",
                        "enum": [
                            "ubuntu-latest",
                            "ubuntu-18.04",
                            "ubuntu-16.04",
                            "windows-latest",
                            "windows-2019",
                            "windows-2016",
                            "macOS-latest",
                            "macOS-10.14",
                        ],
                    },
                    "steps": {
                        "description": "A job contains a sequence of tasks called steps. Steps can run commands, run setup tasks, or run an action in your repository, a public repository, or an action published in a Docker registry. Not all steps run actions, but all actions are run as a step. Each step runs in its own process in the virtual environment and has access to the workspace and filesystem. Because steps are run in their own process, changes to environment variables are not preserved between steps. GitHub provides built-in steps to set up and complete a job.",
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "name": {
                                    "description": "A name for your step to display on GitHub.",
                                    "type": "string",
                                },
                                "uses": {
                                    "description": "Selects an action to run as part of a step in your job. An action is a reusable unit of code. You can use an action defined in the same repository as the workflow, a public repository, or in a published Docker container image.",
                                    "type": "string",
                                },
                                "with": {
                                    "type": "object",
                                    "description": "A map of the input parameters defined by the action. Each input parameter is a key/value pair. Input parameters are set as environment variables. The variable is prefixed with INPUT_ and converted to upper case.",
                                    "addtionalProperties": {"type": "string"},
                                },
                                "env": {
                                    "type": "object",
                                    "description": "Sets environment variables for steps to use in the virtual environment. Public actions may specify expected environment variables in the README file. If you are setting a secret in an environment variable, you must set secrets using the secrets context.",
                                    "addtionalProperties": {"type": "string"},
                                },
                                "if": {
                                    "description": "Identifies any steps that must complete successfully before this step will run. It can be a string or an array of strings. If a step fails, all steps that need it will also fail unless the steps use a conditional statement that causes the step to continue.",
                                    "type": "string",
                                },
                                "run": {
                                    "description": "Runs command line programs using the operating system's shell. If you do not provide a name, the step name will default to the run command. Commands run using non-login shells by default.",
                                    "type": "string",
                                },
                                "working-directory": {
                                    "description": "The default directory that the action uses in a job's workspace.",
                                    "type": "string",
                                },
                                "continue-on-error": {
                                    "description": "Prevents a job from failing when a step fails. Set to true to allow a job to pass when this step fails.",
                                    "type": "boolean",
                                },
                                "timeout-minutes": {
                                    "description": "The maximum number of minutes to let a workflow run before GitHub automatically cancels it.",
                                    "type": "number",
                                },
                            },
                        },
                    },
                },
            },
        },
    },
    "definitions": {"webhook_event": {"enum": ["push"]}},
}

keyvals = {
    "type": "object",
    "patternProperties": {
        "\\w[\\w-]*": {"type": "string"},
    },
}

# We allow updaters to have any dict structure - they are responsible for validating
updaters_schema = {
    "type": "object",
    "patternProperties": {
        "\\w[\\w-]*": {"type": "object"},
    },
}

# Currently all of these are required
settingsProperties = {
    "github_api": {"type": "string"},
    "config_editor": {"type": "string"},
    "updaters": updaters_schema,
    "code_theme": {"type": "string", "choices": list(get_all_styles())},
}

settings = {
    "$schema": schema_url,
    "title": "Settings Schema",
    "type": "object",
    "required": [
        "github_api",
        "config_editor",
        "code_theme",
        "updaters",
    ],
    "properties": settingsProperties,
    "additionalProperties": False,
}



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


import abc
import importlib
import inspect
import os
import re
from collections.abc import Mapping

import jsonschema
import requests

from action_updater.logger import logger

here = os.path.abspath(os.path.dirname(__file__))


class UpdaterFinder(Mapping):
    """
    Create a cache of available updaters.
    """

    _updaters = {}

    def __init__(self):
        """
        Instantiate an updater
        """
        self.collection_path = os.path.join(here, "updaters")
        self.load()

    def __getitem__(self, name):
        return self._updaters.get(name)

    def __iter__(self):
        return iter(self._updaters)

    def __len__(self):
        return len(self._updaters)

    def load(self):
        """
        Load new updaters
        """
        self._updaters = self._load_updaters()

    def _load_updaters(self):
        """
        Load updaters based on listing folders in the collection.
        """
        lookup = {}
        for name in os.listdir(self.collection_path):
            updater_dir = os.path.join(self.collection_path, name)
            updater_file = os.path.join(updater_dir, "update.py")

            # Skip files in collection folder
            if os.path.isfile(updater_dir):
                continue

            # Continue if the file doesn't exist
            if not os.path.exists(updater_file):
                logger.debug("%s does not appear to have an update.py, skipping." % updater_dir)
                continue

            # The class name means we split by underscore, capitalize, and join
            class_name = "".join([x.capitalize() for x in name.split("_")]) + "Updater"
            module = "action_updater.main.updaters.%s.update" % name

            # Not instantiated - will be instantiated for a specific action
            lookup[name] = getattr(importlib.import_module(module), class_name)
        return lookup


class UpdaterBase:

    name = "updater"
    description = "An abstract base updater."
    date_time_format = "%Y-%m-%dT%H:%M:%S%z"

    # The default updater is not intended for static files
    static_files = False

    def __init__(self, token, settings=None):
        self._data = {}
        self.headers = {}
        self.update_token(token)
        self.count = 0

        # Each updater can ship its own settings schema
        if not hasattr(self, "schema"):
            self.schema = {}

        self.validate_settings(settings)

    @abc.abstractmethod
    def detect(self, *args, **kwargs):
        pass

    @property
    def slug(self):
        return re.sub("(-|_)", "", self.name)

    @property
    def title(self):
        return self.name.capitalize()

    def validate_settings(self, settings):
        """
        If settings are provided (and the updater has a schema) ensure we validate.
        """
        self.global_settings = settings or {}
        if self.global_settings and self.schema:
            jsonschema.validate(self.settings, schema=self.schema)

    @property
    def settings(self):
        """
        Get settings specific to updater
        """
        return self.global_settings.updaters.get(self.name, {})

    def update_token(self, token=None):
        """
        Set token and headers, if found
        """
        self.token = token or os.environ.get("GITHUB_TOKEN")
        if self.token:
            self.headers["Authorization"] = "token %s" % self.token

    @property
    def classpath(self):
        return os.path.dirname(inspect.getfile(self.__class__))

    def get_releases(self, repo):
        """
        Get the lateset release of an action (under flux-framework)
        """
        return self.get_request(f"{self.global_settings.github_api}/repos/{repo}/releases")

    def get_tags(self, repo):
        """
        Get the lateset tags for a repository
        """
        return self.get_request(f"{self.global_settings.github_api}/repos/{repo}/git/refs/tags")

    def get_tags_lookup(self, repo):
        """
        This isn't required to be sploot out, but it's easier to debug / read if necessary
        """
        tags = {}
        for t in self.get_tags(repo):
            if "ref" not in t or "refs/tags" not in t["ref"]:
                continue
            tags[re.sub("refs/tags/", "", t["ref"])] = t
        return tags

    def get_request(self, url):
        """
        Perform a GitHub get request (assume pagination)
        """
        response = requests.get(url, headers=self.headers, params={"per_page": 100})

        try:
            response.raise_for_status()
        except Exception:
            # Set a warning about limtis without tokens!
            if not self.token:
                logger.exit("export GITHUB_TOKEN to increase API limits.")

        # latest release should be first in this set
        return response.json()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


from .settings import Settings


def get_client(quiet=False, **kwargs):
    """
    Get an actions updater client.

    This will ensure we add all of the default updaters to our client.
    """
    # TODO we can further customize here.
    validate = kwargs.get("validate", True)

    # Load user settings to add to client, and container technology
    settings = Settings(kwargs.get("settings_file"), validate)

    from .client import ActionUpdater

    ActionUpdater.settings = settings
    client = ActionUpdater(**kwargs, quiet=quiet)
    return client



#!/usr/bin/python

# Copyright (C) 2022 Vanessa Sochat.

# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import shutil

from action_updater.main import get_client

here = os.path.dirname(os.path.abspath(__file__))
root = os.path.dirname(here)


def init_client(tmpdir):
    """
    Get a common client for some container technology and module system
    """
    settings_file = os.path.join(root, "settings.yml")
    new_settings = os.path.join(tmpdir, "settings.yml")
    shutil.copyfile(settings_file, new_settings)
    client = get_client(
        quiet=False,
        settings_file=new_settings,
    )
    return client


def get_updaters():
    client = get_client()
    return client.updaters



#!/usr/bin/env python

__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import argparse
import os
import sys

import action_updater
from action_updater.logger import setup_logger


def get_parser():
    parser = argparse.ArgumentParser(
        description="Actions Updater",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Global Variables
    parser.add_argument(
        "--debug",
        dest="debug",
        help="use verbose logging to debug.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="show software version.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--settings-file",
        dest="settings_file",
        help="custom path to settings file.",
    )

    # On the fly updates to config params
    parser.add_argument(
        "-c",
        dest="config_params",
        help=""""customize a config value on the fly to ADD/SET/REMOVE for a command
action-updater -c set:key:value <command> <args>
action-updater -c add:listkey:value <command> <args>
action-updater -c rm:listkey:value""",
        action="append",
    )

    subparsers = parser.add_subparsers(
        help="actions",
        title="actions",
        description="actions",
        dest="command",
    )

    # print version and exit
    subparsers.add_parser("version", description="show software version")
    subparsers.add_parser("list-updaters", description="list updaters available")

    # Detect changes for a directory or file of interest
    detect = subparsers.add_parser(
        "detect",
        description="detect action updates before applying them.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    update = subparsers.add_parser(
        "update",
        description="detect and apply updates.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    for command in detect, update:
        command.add_argument(
            "paths",
            help="path to run detect over (e.g., yaml file or GitHub actions folder)",
            action="append",
        )
        command.add_argument(
            "-u",
            "--updater",
            dest="updaters",
            help="update using one or more specific updaters",
            action="append",
        )
        command.add_argument(
            "--updaters",
            dest="updater_list",
            help="provide a comma separated value list of updater (e.g., version,setoutput)",
        )
        command.add_argument(
            "--no-details",
            dest="no_details",
            help="do not show details for file changes",
            default=False,
            action="store_true",
        )

    config = subparsers.add_parser(
        "config",
        description="update configuration settings. Use set or get to see or set information.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    config.add_argument(
        "--central",
        dest="central",
        help="make edits to the central config file.",
        default=False,
        action="store_true",
    )

    config.add_argument(
        "params",
        nargs="*",
        help="""Set or get a config value, edit the config, add or remove a list variable, or create a user-specific config.
action-updater config set key value
action-updater config set key:subkey value
action-updater config get key
action-updater edit
action-updater config inituser
action-updater config add registry /tmp/registry
action-updater config remove registry /tmp/registry""",
        type=str,
    )

    return parser


def run_action_updater():

    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client
        and exit with return code.
        """

        version = action_updater.__version__

        print("\nSingularity Registry (HPC) Client v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    if args.debug is True:
        os.environ["MESSAGELEVEL"] = "DEBUG"

    # Show the version and exit
    if args.command == "version" or args.version:
        print(action_updater.__version__)
        sys.exit(0)

    setup_logger(
        quiet=args.quiet,
        debug=args.debug,
    )

    # retrieve subparser (with help) from parser
    helper = None
    subparsers_actions = [
        action for action in parser._actions if isinstance(action, argparse._SubParsersAction)
    ]
    for subparsers_action in subparsers_actions:
        for choice, subparser in subparsers_action.choices.items():
            if choice == args.command:
                helper = subparser
                break

    if args.command == "detect":
        from .detect import main
    elif args.command == "config":
        from .config import main
    elif args.command == "update":
        from .update import main
    elif args.command == "list-updaters":
        from .listing import list_updaters as main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args, parser=parser, extra=extra, subparser=helper)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)


if __name__ == "__main__":
    run_action_updater()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os

from rich.console import Console

import action_updater.utils as utils

from .action import GitHubAction
from .settings import Settings
from .updater import UpdaterFinder


class ActionUpdater:
    """
    Create a GitHub updater
    """

    def __init__(self, quiet=False, token=None, settings_file=None, **kwargs):
        self.token = token
        self._updaters = {}
        self.quiet = quiet
        self.c = Console()

        # If using for a GitHub action, a global flag that indicates changes
        self.has_changes = False

        # If we don't have default settings, load
        if not hasattr(self, "settings"):
            self.settings = Settings(settings_file)

    @property
    def updaters(self):
        """
        Get a list of updaters available
        """
        if not self._updaters:

            # All updaters can be provided with the GitHub token
            self.finder = UpdaterFinder()
            self._updaters = {}
            for name, updaterClass in self.finder.items():

                # Instantiate an updater for the path, provide settings
                self._updaters[name] = updaterClass(token=self.token, settings=self.settings)

        return self._updaters

    def iter_paths(self, paths):
        """
        Helper function to flexibly handle parsing paths.
        """
        # Ensure we start from a list
        if not isinstance(paths, list):
            paths = [paths]

        final = set()

        # Run each updater on each path
        for path in paths:

            if os.path.exists(path) and os.path.isfile(path):
                final.add(path)
                continue
            for filename in utils.recursive_find(path, "[.](yaml|yml)"):
                final.add(filename)
        return list(final)

    def detect(self, paths, details=True, updaters=None):
        """
        Look for changes in files according to updaters
        """
        actions = {}

        for path in self.iter_paths(paths):

            # Load into GitHub action
            action = GitHubAction(path)

            self.c.print(f"⭐️ [yellow]{path}[/yellow]")

            # Todo convert this into an iter function (shared between detect and update)
            for _, updater in self.updaters.items():

                # Skip updaters per request of the user
                if updaters and updater.slug not in updaters:
                    continue

                # The count reflects the last run
                if updater.detect(action):
                    self.c.print(f"[red]✖️ {updater.title} Updater: {updater.count} updates[/red]")
                    self.has_changes = True
                else:
                    self.c.print(f"[green]✔ {updater.title}: No updates[/green]")

            # If we want to show details:
            if details:
                action.diff(self.settings.code_theme or "vim")
            actions[path] = action
        return actions

    def update(self, paths, details=True, updaters=None):
        """
        Update files.
        """
        actions = self.detect(paths, details=details, updaters=updaters)
        for path, action in actions.items():
            if action.has_changes:
                self.c.print(f"[purple]❇ Writing updated {path}[/purple]")
                action.write(path)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[action-updater]"



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os

import action_updater.utils as utils

install_dir = utils.get_installdir()
reps = {"$install_dir": install_dir, "$root_dir": os.path.dirname(install_dir)}

# The default settings file in the install root
default_settings_file = os.path.join(reps["$install_dir"], "settings.yml")

# The user settings file can be created to over-ride default
user_settings_file = os.path.join(os.path.expanduser("~/.action-updater"), "settings.yml")



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


import copy
import difflib

from rich.console import Console
from rich.markdown import Markdown

import action_updater.utils as utils


class GitHubAction:
    """
    Parse a GitHub action into it's sections.

    We always present the changes (copy of original) and then can
    easily compare the two. The overall structure should not change.
    """

    def __init__(self, filename):
        self.cfg = utils.read_yaml(filename)
        self.changes = copy.deepcopy(self.cfg)

    @property
    def jobs(self):
        return self.changes.get("jobs")

    def write(self, path):
        """
        Save the action to file.
        """
        utils.write_yaml(self.changes, path)

    @property
    def has_changes(self):
        """
        Determine if before != after (the action has changed)
        """
        return self.render_before() != self.render_after()

    def render_after(self):
        """
        Render the action post-detect (with changes).
        """
        return utils.get_yaml_string(self.changes).splitlines(keepends=True)

    def render_before(self):
        """
        Render the action pre-detect (without changes).
        """
        return utils.get_yaml_string(self.cfg).splitlines(keepends=True)

    def diff(self, code_theme="vim"):
        """
        Show diff between original (cfg) and changed!
        """
        before = self.render_before()
        after = self.render_after()

        if before == after:
            print()
            return

        diff = "".join(
            list(
                difflib.unified_diff(
                    before,
                    after,
                    "original",
                    "updated",
                )
            )
        )

        c = Console()
        md = Markdown(f"""\n```diff\n{diff}\n```\n""", code_theme=code_theme)
        c.print(md)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import re

from action_updater.main.updater import UpdaterBase


def update_lines(lines, key="set-output", envar="$GITHUB_OUTPUT"):
    """
    Helper function to replace generic "set-X" variable with pipe to envar.
    """
    lines = lines.splitlines(keepends=True)

    updated = []

    for line in lines:
        if key not in line:
            updated.append(line)
            continue

        # Convert:
        # echo "::set-output name={name}}::${raw_data[4]}"
        # echo "name=value" >> $GITHUB_OUTPUT
        match = re.search("%s(\s)+name=(?P<name>.+)::(?P<value>.+)" % key, line)  # noqa
        if not match:
            updated.append(line)
            continue

        match = match.groupdict()
        value = match["value"].strip('"').strip("'")
        line = 'echo "%s=%s" >> %s\n' % (match["name"], value, envar)
        updated.append(line)

    return "".join(updated)


class SetoutputUpdater(UpdaterBase):

    name = "set-output"
    description = "update deprecated set-output commands"

    def detect(self, action):
        """
        Detect changes in an action, old set-output.
        """
        # Set the count to 0
        self.count = 0

        # No point if we don't have jobs!
        if not action.jobs:
            return False

        # For each job, look for steps->updater versions
        for _, job in action.jobs.items():
            for step in job.get("steps", []):

                # We are primarily interested in uses
                if "run" not in step:
                    continue

                # Update step run lines (returns parsed again together)
                updated_lines = update_lines(step["run"])

                # Keep track of change counts
                if updated_lines != step["run"]:
                    self.count += 1

                step["run"] = updated_lines

        return self.count != 0



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from action_updater.main.updater import UpdaterBase
from action_updater.main.updaters.setoutput.update import update_lines


class SavestateUpdater(UpdaterBase):

    name = "save-state"
    description = "update deprecated save-state commands"

    def detect(self, action):
        """
        Detect changes in an action, old set-state.
        """
        # Set the count to 0
        self.count = 0

        # No point if we don't have jobs!
        if not action.jobs:
            return False

        # For each job, look for steps->updater versions
        for _, job in action.jobs.items():
            for step in job.get("steps", []):

                # We are primarily interested in uses
                if "run" not in step:
                    continue

                # Update step run lines
                updated_lines = update_lines(step["run"], "save-state", "$GITHUB_STATE")

                # Keep track of change counts
                if updated_lines != step["run"]:
                    self.count += 1

                step["run"] = updated_lines

        return self.count != 0



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import inspect
import logging as _logging
import os
import platform
import random
import sys
import threading

import rich.console
import rich.table


class Table:
    """
    Format a list of dicts into a table
    """

    def __init__(self, items):
        self.data = items
        self.max_widths = {}
        self.ensure_complete()

    def available_width(self, columns):
        """
        Calculate available width based on fields we cannot truncate (urls)
        """
        # We will determine column width based on terminal size
        try:
            width = os.get_terminal_size().columns
        except OSError:
            width = 120

        # Calculate column width
        column_width = int(width / len(columns))
        updated = width

        for _, needed in self.max_widths.items():
            updated = updated - needed

        # We don't have enough space
        if updated < 0:
            logger.warning("Terminal is too small to correctly render!")
            return column_width

        # Otherwise, recalculate column width taking into account truncation
        # We use the updated smaller width, and break it up between columns
        # that don't have a max width
        return int(updated / (len(columns) - len(self.max_widths)))

    def ensure_complete(self):
        """
        If any data missing fields, ensure they are included
        """
        fields = set()
        for entry in self.data:
            [fields.add(x) for x in entry.keys()]

        # Ensure fields are present
        for entry in self.data:
            for field in fields:

                if field not in entry:
                    entry[field] = ""

    @property
    def color(self):
        """
        Return a random color
        """
        return "color(" + str(random.choice(range(255))) + ")"

    def table_columns(self):
        """
        Shared function to return consistent table columns
        """
        if not self.data:
            return []
        return list(self.data[0].keys())

    def table_rows(self, columns, limit=25):
        """
        Shared function to yield rows as a list
        """
        # All keys are lowercase
        column_width = self.available_width(columns)

        for i, row in enumerate(self.data):

            # have we gone over the limit?
            if limit and i > limit:
                return

            parsed = []
            for column in columns:
                content = row[column]
                if (
                    content
                    and len(content) > column_width
                    and column not in self.endpoint.truncate_list
                ):
                    content = content[:column_width] + "..."
                parsed.append(content)
            yield parsed

    def show(self, limit=25, title=None):
        """
        Pretty print a table of content
        """
        table = rich.table.Table(title=title)

        # Always skip these columns
        seen_colors = []

        # Get column titles
        columns = self.table_columns()
        for column in columns:
            color = None
            while not color or color in seen_colors:
                color = self.color
            table.add_column(column.capitalize(), style=color)

        # Add rows
        for row in self.table_rows(columns, limit=limit):
            table.add_row(*row)

        # And print!
        console = rich.console.Console()
        console.print(table, justify="left")


class LogColors:
    PURPLE = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    RED = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def underline(msg):
    """
    Return an underlined message
    """
    return f"{LogColors.UNDERLINE}{msg}{LogColors.ENDC}"


def add_prefix(msg, char=">>"):
    """
    Add an "OKBLUE" prefix to a message
    """
    return f"{LogColors.OKBLUE}{char}{LogColors.ENDC} {msg}"


class ColorizingStreamHandler(_logging.StreamHandler):

    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
    RESET_SEQ = LogColors.ENDC
    COLOR_SEQ = "\033[%dm"
    BOLD_SEQ = "\033[1m"

    colors = {
        "WARNING": YELLOW,
        "INFO": GREEN,
        "DEBUG": BLUE,
        "CRITICAL": RED,
        "ERROR": RED,
    }

    def __init__(self, nocolor=False, stream=sys.stderr, use_threads=False):
        super().__init__(stream=stream)
        self._output_lock = threading.Lock()
        self.nocolor = nocolor or not self.can_color_tty()

    def can_color_tty(self):
        if "TERM" in os.environ and os.environ["TERM"] == "dumb":
            return False
        return self.is_tty and not platform.system() == "Windows"

    @property
    def is_tty(self):
        isatty = getattr(self.stream, "isatty", None)
        return isatty and isatty()

    def emit(self, record):
        with self._output_lock:
            try:
                self.format(record)  # add the message to the record
                self.stream.write(self.decorate(record))
                self.stream.write(getattr(self, "terminator", "\n"))
                self.flush()
            except BrokenPipeError as e:
                raise e
            except (KeyboardInterrupt, SystemExit):
                # ignore any exceptions in these cases as any relevant messages have been printed before
                pass
            except Exception:
                self.handleError(record)

    def decorate(self, record):
        message = record.message
        message = [message]
        if not self.nocolor and record.levelname in self.colors:
            message.insert(0, self.COLOR_SEQ % (30 + self.colors[record.levelname]))
            message.append(self.RESET_SEQ)
        return "".join(message)


class Logger:
    def __init__(self):
        self.logger = _logging.getLogger(__name__)
        self.log_handler = [self.text_handler]
        self.stream_handler = None
        self.printshellcmds = False
        self.quiet = False
        self.logfile = None
        self.last_msg_was_job_info = False
        self.logfile_handler = None

    def cleanup(self):
        if self.logfile_handler is not None:
            self.logger.removeHandler(self.logfile_handler)
            self.logfile_handler.close()
        self.log_handler = [self.text_handler]

    def handler(self, msg):
        for handler in self.log_handler:
            handler(msg)

    def set_stream_handler(self, stream_handler):
        if self.stream_handler is not None:
            self.logger.removeHandler(self.stream_handler)
        self.stream_handler = stream_handler
        self.logger.addHandler(stream_handler)

    def set_level(self, level):
        self.logger.setLevel(level)

    def location(self, msg):
        callerframerecord = inspect.stack()[1]
        frame = callerframerecord[0]
        info = inspect.getframeinfo(frame)
        self.debug("{}: {info.filename}, {info.function}, {info.lineno}".format(msg, info=info))

    def yellow(self, msg):
        self.handler(dict(level="info", msg=msg))

    def info(self, msg):
        self.handler(dict(level="info", msg=msg))

    def warning(self, msg):
        self.handler(dict(level="warning", msg=msg))

    def debug(self, msg):
        self.handler(dict(level="debug", msg=msg))

    def error(self, msg):
        self.handler(dict(level="error", msg=msg))

    def exit(self, msg, return_code=1):
        self.handler(dict(level="error", msg=msg))
        sys.exit(return_code)

    def progress(self, done=None, total=None):
        self.handler(dict(level="progress", done=done, total=total))

    def shellcmd(self, msg):
        if msg is not None:
            msg = dict(level="shellcmd", msg=msg)
            self.handler(msg)

    def text_handler(self, msg):
        """The default log handler prints the output to the console.
        Args:
            msg (dict):     the log message dictionary
        """
        level = msg["level"]
        if level == "info" and not self.quiet:
            self.logger.info(msg["msg"])
        if level == "warning":
            self.logger.warning(msg["msg"])
        elif level == "error":
            self.logger.error(msg["msg"])
        elif level == "debug":
            self.logger.debug(msg["msg"])
        elif level == "progress" and not self.quiet:
            done = msg["done"]
            total = msg["total"]
            p = done / total
            percent_fmt = ("{:.2%}" if p < 0.01 else "{:.0%}").format(p)
            self.logger.info("{} of {} steps ({}) done".format(done, total, percent_fmt))
        elif level == "shellcmd":
            if self.printshellcmds:
                self.logger.warning(msg["msg"])


logger = Logger()


def setup_logger(
    quiet=False,
    printshellcmds=False,
    nocolor=False,
    stdout=False,
    debug=False,
    use_threads=False,
    wms_monitor=None,
):
    # console output only if no custom logger was specified
    stream_handler = ColorizingStreamHandler(
        nocolor=nocolor,
        stream=sys.stdout if stdout else sys.stderr,
        use_threads=use_threads,
    )
    logger.set_stream_handler(stream_handler)
    logger.set_level(_logging.DEBUG if debug else _logging.INFO)
    logger.quiet = quiet
    logger.printshellcmds = printshellcmds



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from action_updater.main.github import sort_major, sort_tags
from action_updater.main.updater import UpdaterBase

schema = {
    "type": "object",
    "properties": {
        # Allow these orgs to use major version strings
        "major_orgs": {"type": "array", "items": {"type": "string"}}
    },
    "additionalProperties": False,
}


class VersionUpdater(UpdaterBase):

    name = "version"
    description = "update action versions"
    schema = schema
    cache = {"tags": {}}

    def detect(self, action):
        """
        Detect changes (to be later saved)
        """
        # Set the count to 0
        self.count = 0

        # No point if we don't have jobs!
        if not action.jobs:
            return False

        # We will use major versions for these orgs (trusted)
        trusted_orgs = self.settings.get("major_orgs")

        # For each job, look for steps->updater versions
        for _, job in action.jobs.items():
            for step in job.get("steps", []):

                # We are primarily interested in uses
                if "uses" not in step:
                    continue

                repo = step["uses"]

                # If we have a local action, nothing to update
                if repo.startswith("./"):
                    continue

                # Get the current tag or version (we will want to maintain this convention)
                # Create a lookup based on the ref
                repo, _ = repo.split("@")
                org, _ = repo.split("/", 1)

                # Retrieve all tags for the repository, a lookup by tag name
                tags = self.cache["tags"].get(repo) or self.get_tags_lookup(repo)

                updated = None
                if trusted_orgs and org in trusted_orgs:
                    updated = self.get_major_tag(tags)

                if not updated:
                    updated = self.get_tagged_commit(tags)

                if not updated:
                    continue
                updated = f"{repo}@{updated}"
                previous = step["uses"]

                # If we added a new comment, update the old one
                if "#" in updated:
                    updated, comment = updated.split("#", 1)
                    comment = comment.strip()
                    step["uses"] = updated.strip()

                    # TODO some check to preserve other previous comments?
                    step.ca.items["uses"] = [None, None, None, None]

                    # Add the end of line comment (third position in list)
                    step.yaml_add_eol_comment(f"# {comment}\n", "uses", column=0)

                # Always do the update (regardless of comment!)
                step["uses"] = updated.strip()

                # Do we have a change?
                if step["uses"] != previous:
                    self.count += 1

        return self.count != 0

    def get_major_tag(self, tags):
        """
        Given a list of repository tags, get the most up-todate!
        """
        # If we have a major org, we trust it and want the major version tag
        tags_list = list(tags)
        tags_list = [x for x in tags_list if "." not in x]

        # Cut out early if we don't have major-only tags
        if not tags_list:
            return []

        # Get ordered major tags
        ordered = sort_major(tags_list)

        # If we don't have ordered, we could use branches, but skip for now
        if not ordered:
            return []

        tag = ordered[0]._original

        # Use latest release verbatim
        return tag

    def get_tagged_commit(self, tags):
        """
        Given a list of repository tags, get the most up-todate!
        """
        # Get ordered major tags
        ordered = sort_tags(list(tags))

        # If we don't have ordered, we could use branches, but skip for now
        if not ordered:
            return []

        tag = ordered[0]._original
        meta = tags[tag]
        commit = meta["object"]["sha"]
        return f"{commit} # {tag}"



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


def parse_updaters(args):
    """
    Parse a unique set of updaters from --update and --updaters
    """
    updaters = args.updaters or []
    if args.updater_list:
        updaters += [x.strip() for x in args.updater_list.split(",") if x.strip()]
    return list(set(updaters))



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from action_updater.logger import logger
from action_updater.main import get_client

from .helpers import parse_updaters


def main(args, parser, extra, subparser):
    cli = get_client(quiet=args.quiet)
    cli.detect(paths=args.paths, details=not args.no_details, updaters=parse_updaters(args))
    if cli.has_changes:
        logger.exit("Found changes, exiting with non-zero code.")
